# using Click implementation
import re
import subprocess

from pathlib import Path
from typing import List, Optional, Tuple

import click  # CLI magic

from todotree.Addons import Addons
from todotree.Errors.TodoFileNotFound import TodoFileNotFound
from todotree.Errors.DoneFileNotFound import DoneFileNotFound
from todotree.Task.DoneTask import DoneTask
from todotree.Taskmanager import Taskmanager
from todotree.Config import Config
from todotree.MainUtils import MainUtils

# GLOBALS. #
config = Config()
addons = Addons(config)
taskManager = Taskmanager(config)


# Click Replaces:
# Argparse
#
# NOTE: this file cannot be in a class. See: https://github.com/pallets/click/issues/601
# But context and variable ferrying can be done using the context option.
# We just call the context 'self' and hope the issue does resolve itself.
# https://click.palletsprojects.com/en/8.1.x/commands/#nested-handling-and-contexts


@click.group()
@click.option('--config-path', default=None, help="Path to the configuration file.")
@click.option('--todo-file', default=None, help="Path to the todo.txt file, overrides --config.")
@click.option('--done-file', default=None, help="Path to the done.txt file, overrides --config.")
@click.option('--verbose', is_flag=True, help="Increases verbosity in messages.", is_eager=True)
@click.option('--quiet', is_flag=True, help="Do not print messages, only output. Useful in scripts.",
              is_eager=True)
def root(config_path: Optional[Path], todo_file: Optional[Path], done_file: Optional[Path],
         verbose: bool, quiet: bool):
    """
    The main list of todotree's command.

    Options must be before the actual argument,
    so `todotree --verbose list` will work, but `todotree list --verbose` will not.
    """
    # ^ This text also shows up in the help command.
    # Root click group. This manages all the command line interactions.
    # ensure that ctx.obj exists and is a dict
    config.verbose = False  # FUTURE: Fix properly.
    MainUtils.initialize(config_path, done_file, quiet, config, todo_file, verbose)
    # Pass to another command.
    pass


@root.command('add', short_help='Add a task to the task list')
@click.argument('task', type=str, nargs=-1)
def add(task: Tuple):
    # Convert tuple to string
    task: str = " ".join(map(str, task))
    try:
        # Disable fancy imports, because they are not needed.
        config.enable_project_folder = False
        # Import tasks.
        taskManager.import_tasks()
        # Add task
        new_number = taskManager.add_task_to_file(task.strip() + "\n")
        config.console.info("Task added:")
        click.echo(f"{new_number} {task}")
        MainUtils.commit_exit("add", config)
    except TodoFileNotFound as e:
        MainUtils.handle_todo_file_not_found(e, config)
        exit(1)


@root.command('addx', short_help='Add a task and immediately mark it as done')
@click.argument('task', type=str, nargs=-1)
def add_x(task: Tuple):
    """
    Adds a completed task to done.txt. The task is not added to todo.txt.
    :param task: The task to add to done.txt.
    """
    try:
        done = DoneTask.task_to_done(" ".join(map(str, task)))
        with config.done_file.open("a") as f:
            f.write(done)
        click.echo(done)
    except FileNotFoundError as e:
        MainUtils.handle_done_file_not_found(e, config)
        exit(1)


@root.command('append', short_help='append `append_string` to `task_nr`')
@click.argument('task_nr', type=int)
@click.argument('append_string', type=str, nargs=-1)
def append(task_nr: int, append_string: Tuple[str]):
    """
    Appends the contents of append_string to the task represented by task_nr.
    A space is inserted between the two tasks, so you do not have to worry that words aretogether.
    """
    # Disable fancy imports, because they are not needed.
    config.enable_project_folder = False
    # Import tasks.
    try:
        taskManager.import_tasks()
    except TodoFileNotFound as e:
        MainUtils.handle_todo_file_not_found(e, config)
        exit(1)
    # Convert tuple to string.
    append_string = " ".join(append_string)

    # Append task.
    config.console.info("The new task is: ")
    click.echo(taskManager.append_to_task(task_nr, append_string.strip()))
    MainUtils.commit_exit("append", config)


@root.command('context', short_help='list task in a tree, by context')
def context():
    """list a tree, of which the first node is the context, the second nodes are the tasks"""
    try:
        taskManager.import_tasks()
    except TodoFileNotFound as e:
        MainUtils.handle_todo_file_not_found(e, config)
        exit(1)
    click.echo(taskManager.print_context_tree())
    exit()


@root.command('cd', short_help='print directory of the todo.txt directory')
def cd():
    """print directory of the todo.txt directory"""
    config_path: Path = Path(config.todo_folder)
    if config.verbose:
        config.console.info("The location to the data folder is: ")

    if config_path.is_absolute():
        # Then the configured path is printed.
        click.echo(str(config_path))
        exit()
    # Then the relative path is resolved to be absolute.
    base_path: Path = Path.home()
    full_path: Path = base_path / config_path
    click.echo(str(full_path))
    exit()


@root.command('do', short_help='mark task as done and move it to the done.txt')
@click.argument('task_numbers', type=list, nargs=-1)  # type=list[int]
def do(task_numbers: List[Tuple]):
    """
    Mark tasks as done, therefor moving them to done.txt with a date stamp of today.
    :param task_numbers: The list of tasks which are completed.
    """
    # Convert to ints. Task numbers is a list of tuples. Each tuple contains one digit of the number.
    new_numbers: List[int] = []
    for task_tuple in task_numbers:
        new_number: str = ""
        for task_digit in task_tuple:
            new_number += task_digit
        new_numbers.append(int(new_number))
    # Write back to old value.
    task_numbers = new_numbers
    # Marking something as Done cannot be done with fancy imports
    # So we disable them.
    config.enable_project_folder = False
    try:
        taskManager.import_tasks()
    except TodoFileNotFound as e:
        MainUtils.handle_todo_file_not_found(e, config)
        exit(1)
    try:
        completed_tasks = taskManager.mark_as_done(task_numbers)
    except DoneFileNotFound as e:
        MainUtils.handle_done_file_not_found(e, config)
        exit(1)
    # Print the results
    config.console.info("Tasks marked as done:")
    for task in completed_tasks:
        click.echo(task)
    MainUtils.commit_exit("do", config)


@root.command('due', short_help='List tasks by their due date')
def due():
    """List tasks in a tree by their due date."""
    # Disable fancy imports, because they do not have due dates.
    config.enable_project_folder = False
    # Import tasks.
    try:
        taskManager.import_tasks()
    except TodoFileNotFound as e:
        MainUtils.handle_todo_file_not_found(e, config)
        exit(1)
    # Print due tree.
    click.echo(taskManager.print_by_due())
    exit()


@root.command('edit', short_help='open the todo.txt in an editor.')
def edit():
    """
    Open the todo.txt in an editor for manual editing of tasks.
    This is useful when you need to modify a lot of tasks, which would be complicated when doing it with todotree.
    """
    # Disable fancy imports.
    config.enable_project_folder = False
    click.edit(filename=str(config.todo_file))
    MainUtils.commit_exit("edit", config)


@root.command('filter', short_help='only show tasks containing the search term.')
@click.argument('search_term')
def filter_list(search_term):
    """
    Only show tasks which have search term in them. This can also be a keyword.

    :param search_term: The term to search.
    """
    try:
        taskManager.import_tasks()
    except TodoFileNotFound as e:
        MainUtils.handle_todo_file_not_found(e, config)
        exit(1)

    if config.verbose:
        config.console.info("The todo list is:")
    elif not config.quiet:
        config.console.info("Todos")

    taskManager.filter_by_string(search_term)
    click.echo(taskManager)


@root.command('list', short_help='List tasks')
def list_tasks():
    """
    Print a flat list of tasks, sorted by their priority.
    """
    try:
        taskManager.import_tasks()
    except TodoFileNotFound as e:
        MainUtils.handle_todo_file_not_found(e, config)
        exit(1)

    if config.verbose:
        config.console.info("The todo list is:")
    elif not config.quiet:
        config.console.info("Todos")

    click.echo(taskManager)


@root.command('list_done', short_help='List tasks which are marked as done')
def list_done():
    """
    List tasks which are marked as done. The numbers can be used with the revive command.
    """
    try:
        taskManager.list_done()
    except DoneFileNotFound as e:
        MainUtils.handle_done_file_not_found(e, config)
        exit(1)


@root.command('print_raw', short_help='print todo.txt without any formatting or filtering')
def print_raw():
    """
    Output the todo.txt without any processing.
    This is equivalent to `cat $(todo cd)/todo.txt` in bash.
    """
    try:
        with open(taskManager.config.todo_file, "r") as f:
            click.echo(f.read())
    except FileNotFoundError as e:
        MainUtils.handle_todo_file_not_found(e, config)


@root.command('priority', short_help='set new priority to task')
@click.argument('task_number', type=int)
@click.argument('new_priority', type=str)
def priority(task_number, new_priority):
    """
    Adds or updates the priority of the task.
    :param task_number: The task to re-prioritize.
    :param new_priority: The new priority.
    """
    # Disable fancy imports.
    config.enable_project_folder = False
    # Run task.
    try:
        taskManager.import_tasks()
    except TodoFileNotFound as e:
        MainUtils.handle_todo_file_not_found(e, config)
        exit(1)
    taskManager.add_or_update_priority(
        priority=(new_priority.upper()), task_number=task_number)


@root.command('project', short_help='print tree by project')
def project():
    # Import tasks.
    try:
        taskManager.import_tasks()
    except TodoFileNotFound as e:
        MainUtils.handle_todo_file_not_found(e, config)
        exit(1)
    # Print due tree.
    click.echo(taskManager.print_project_tree())
    exit()


@root.command('revive', short_help='Revive a task that was accidentally marked as done.')
@click.argument('done_number', type=int)
def revive(done_number):
    """
    Move a task from done.txt to todo.txt.
    The `done_number` can be found using the `list_done` command.
    :param done_number: The number of the task to revive.
    """
    try:
        click.echo(taskManager.revive_task(done_number))
    except TodoFileNotFound as e:
        MainUtils.handle_todo_file_not_found(e, config)
        exit(1)
    except DoneFileNotFound as e:
        MainUtils.handle_done_file_not_found(e, config)
        exit(1)
    MainUtils.commit_exit("revive", config)


@root.command('schedule', short_help='hide task until date.')
@click.argument('task_number', type=int)
@click.argument('new_date', type=str, nargs=-1)
def schedule(task_number: int, new_date: Tuple[str]):
    """
    hide the task until the date given. If new_date is not in ISO format (yyyy-mm-dd) such as "Next Wednesday",
    then it tries to figure out the date with the `date` program, which is only in linux.
    """
    # Disable fancy imports, because they do not have t dates.
    config.enable_project_folder = False
    # Convert
    date = " ".join(new_date)
    date_pattern = re.compile(r"(\d{4})-(\d{2})-(\d{2})")
    if not date_pattern.match(date):
        if config.verbose:
            config.console.info(f"Attempt to parse {date} with the `date` program.")
        # Try to use the `date` utility.
        dat = subprocess.run(
            ["date", "-d " + date, "+%F "],
            capture_output=True,
            encoding="utf-8"
        )
        if dat.returncode > 0:
            config.console.error(f"The date {new_date} could not be parsed.")
            exit(1)
        date = dat.stdout.strip()
    try:
        taskManager.import_tasks()
        config.console.info(f"Task {task_number} hidden until {date}")
        updated_task = taskManager.add_or_update_t_date(date, task_number)
        config.console.info(str(updated_task))
    except TodoFileNotFound as e:
        MainUtils.handle_todo_file_not_found(e, config)
        exit(1)

    MainUtils.commit_exit("schedule", config)


@root.command('addons', short_help='Run an addon script')
@click.argument('command', type=str)
def addons_command(command: str):
    """
    Run an addon script.
    The script must be in the addons_folder. It can be any language you like: It does not have to be python.
    However, it must have the executable bit set.

    :param command: The script/command to run.
    """
    try:
        result = Addons(config).run(command)
    except FileNotFoundError:
        config.console.error(f"There is no script at {Path(config.addons_folder / command)}")
        exit(1)
    click.echo(result.stdout, nl=False)
    if result.returncode != 0:
        click.echo(result.stderr)
    MainUtils.commit_exit(f"addons {command}", config)


#  End Region Command Definitions.
#  Setup Click

CONTEXT_SETTINGS: dict = dict(help_option_names=['-h', '--help'])
"""Click context settings. See https://click.palletsprojects.com/en/8.1.x/complex/ for more information."""
cli: click.CommandCollection = click.CommandCollection(
    sources=[root],
    context_settings=CONTEXT_SETTINGS
)
"""Command Collection defining defaults. https://click.palletsprojects.com/en/8.1.x/api/#click.CommandCollection ."""

if __name__ == '__main__':
    cli()



import click


class ConsolePrefixes:
    """Class for handling the prefixes in the terminal.

    The class is somewhat similar to a simple logger. It prints messages with a configured prefix.
    """

    def __init__(self, enable_colors: bool, console_good: str, console_warn: str, console_error: str):
        """
        Initialize a new Console Prefix.
        :param enable_colors: Whether to enable colors.
        :param console_good: The prefix for information messages.
        :param console_warn: The prefix for warning messages.
        :param console_error: The prefix for error messages.
        """
        self.enable_colors = enable_colors
        self.console_error = console_error
        self.console_warn = console_warn
        self.console_good = console_good

    def info(self, message):
        """
        Print an information message to console.
        :param message: The message.
        """
        self.__emit(self.console_good, 'green', message)

    def warning(self, message: str):
        """
        Print a warning message to the console.
        :param message: The message
        """
        self.__emit(self.console_warn, 'yellow', message)

    def error(self, message: str):
        """
        Print an error message to the console.
        :param message: The message to print.
        """
        self.__emit(self.console_error, 'red', message)

    def __emit(self, prefix: str, color: str, message: str):
        """
        Print a message with a prefix.
        :param prefix: The prefix of the message.
        :param message: The message itself.
        """
        # Emit the prefix.
        if self.enable_colors:
            click.secho(prefix, fg=color, nl=False)
        else:
            click.echo(prefix, nl=False)
        # Emit the message.
        click.echo(message)



from todotree.Task.Task import Task


class FileTask(Task):
    """
    Task class for printing to a file.
    """

    def __init__(self, i: int, task_string: str):
        super().__init__(i, task_string)



from typing import Dict, Callable

from todotree.Task.Task import Task
from todotree.Config import Config


class Tree:
    """
    Class representing a tree.

    The tree is a graph structure where there are no loops.
    However, this one is a special version of it.
    This tree is only 2 nodes deep.
    The leaves are the actual objects.
    The first layer of nodes consist of the value of one of the attributes of the leaves.
    Leaves with the same attribute are connected to the same node.

    To print the tree, simply cast it to string, like so `str(Tree())`.
    """

    def __init__(self, list_to_convert, key, **kwargs):
        """
        :param list_to_convert: the list to convert to a tree.
        :param key: the key of the first node in the tree.

        :param Optional arguments:
          - root_name: The display name of the root node.
          - empty_string: The display name of the "empty" or "default" first node.
          - config: The configuration file for printing the tree.
        """

        self.data_structure: Dict[str, list[Task]] = {}
        """
        The data structure of this tree.
        """

        self.root_name: str = kwargs.get("root_name", "root")
        """
        The name of the root of the tree.
        """

        self.empty_string: str = kwargs.get("empty_string", "default")
        """
        The 'empty' category.
        """

        self.config: Config = kwargs.get('config', Config())
        """
        The application configuration.
        """

        self.print_func: Callable[[str, Task], str] = kwargs.get('print_func', lambda key_value, task: str(task))
        """
        Function which prints the task.
        Input are: key_value: The value of the key given by `key`.
        Output must be a string.
        """

        # Generate tree.
        self.data_structure = self.generate_tree(list_to_convert, key)

    @staticmethod
    def dict_add(dictionary, key, value):
        """
        Appends the item to the key if key exists, else creates the key with the item.
        :return:
        :param dictionary: The dictionary to mutate.
        :param key: The key.
        :param value: The value.
        :return: The mutated dictionary.
        """
        if key in dictionary:
            dictionary[key].append(value)
        else:
            dictionary[key] = [value]
        return dictionary

    def generate_tree(self, list_to_parse: list, key: str) -> dict:
        """
        Generate a tree like dictionary from the task list.
        The first line is each distinct key.
        Which points to a list of tasks.

        :key: The key to index to, such as due:date, project or context.
        :return: the generated Dictionary.
        """
        result_dict = {}
        for task in list_to_parse:
            task_attribute = getattr(task, key)
            if isinstance(task_attribute, list):
                if len(task_attribute) == 0:
                    self.dict_add(result_dict, self.empty_string, task)
                else:
                    for key_value in task_attribute:
                        self.dict_add(result_dict, key_value, task)
            else:
                # We are not in a list situation.
                if task_attribute:
                    # Then add the attribute normally.
                    self.dict_add(result_dict, task_attribute, task)
                else:
                    # Then the task attribute value is falsy, add it to the default category.
                    self.dict_add(result_dict, self.empty_string, task)
        return result_dict

    def __str__(self):
        """
        Used for pretty printing the tree.
        """
        # Edge case: empty list -> empty string.
        if self.data_structure == {}:
            return ""

        return_string = self.root_name + "\n"
        # Enumerate to determine whether we are at the end of the list.
        len_first_node = len(self.data_structure)
        for i, first_node in enumerate(self.data_structure.keys()):
            first_list = self.data_structure[first_node]

            # Print the first node. This prints the value of the key.
            return_string += self.config.t_print if (i < len_first_node - 1) else self.config.s_print
            # Section with the name of the first node.
            return_string += first_node + "\n"

            # Inner loop.
            first_line = self.config.l_print if (i < len_first_node - 1) else self.config.e_print
            len_second_node = len(first_list)
            for j, second_node in enumerate(first_list):
                # First node end of list
                second_line = self.config.t_print if (j < len_second_node - 1) else self.config.s_print
                # Line with the actual task.
                return_string += first_line + second_line + self.print_func(first_node, second_node) + "\n"

        return return_string



from datetime import datetime, timedelta
from importlib.metadata import version
from pathlib import Path
from typing import Optional

import click
from git import Repo

from todotree.ConsolePrefixes import ConsolePrefixes
from todotree.Errors.ConfigFileNotFound import ConfigFileNotFound
from todotree.Config import Config


class MainUtils:
    """
    Class containing static methods which are useful in the main class.
    """

    @staticmethod
    def initialize(config_path: Optional[Path], done_file: Optional[Path], quiet: bool, config: Config,
                   todo_file: Optional[Path], verbose: bool):
        """
        Initializes the application and loads the configuration.
        :param config: The global configuration to modify.
        :param config_path: Path to the configuration file from the command line.
        :param done_file: Path to the configuration file from the command line.
        :param quiet: --quiet value
        :param todo_file: Path to the configuration file from the command line.
        :param verbose: --verbose value.
        """
        # parsing arguments.
        try:
            config.read(config_path)
        except ConfigFileNotFound as e:
            MainUtils.handle_config_file_not_found(e, config_path, verbose)

        if todo_file is not None:
            config.todo_file = Path(todo_file)
        if done_file is not None:
            config.done_file = Path(done_file)
        if verbose:
            config.verbose = True
        if quiet:
            config.quiet = True
            config.verbose = False
        # Git pull (if needed).
        MainUtils.git_pull(config)
        # Logging
        if config.verbose:
            config.console.info(f"Version of TodoTree: {version('todotree')}")
            config.console.info(f"Read configuration from {config.config_file}")
            config.console.info(f"The todo file is supposed to be at {config.todo_file}")
            config.console.info(f"The done file is supposed to be at {config.done_file}")

    @staticmethod
    def handle_config_file_not_found(e: BaseException, config_path: Optional[Path], verbose: bool):
        """Handle when the configuration file is not found.

        :param verbose: Whether to include additional information.
        :param config_path: The path where the file was not found.
        :param e: The exception raised.
        """
        # Gentoo style prefixes.
        cp = ConsolePrefixes(True, " * ", " * ", " * ")
        cp.warning(f"The config.yaml file could not be found at {config_path}.")
        if verbose:
            cp.warning(f"The config file should be located at {config_path}")
            cp.warning(str(e))
        cp.warning("The default options are now used.")

    @staticmethod
    def handle_todo_file_not_found(e: BaseException, config: Config):
        """Inform the user that the todo.txt was not found."""
        config.console.error("The todo.txt could not be found.")
        config.console.error(f"It searched at the following location: {config.todo_file}")
        if config.verbose:
            config.console.error(str(e))

    @staticmethod
    def handle_done_file_not_found(e: BaseException, config: Config):
        """Inform the user that the done.txt was not found."""
        config.console.error("The done.txt could not be found.")
        config.console.error(f"It searched at the following location: {config.done_file}")
        if config.verbose:
            click.echo(e)

    @staticmethod
    def commit_exit(action: str, config: Config):
        """
        Commit the files with git before exiting.

        :param config: The configuration.
        :param action: The name of the action, such as list or add.
        """
        if config.git_mode not in ["Local", "Full"]:
            exit()
        repo = Repo(config.todo_folder)

        if repo.is_dirty():
            # Git add.
            repo.index.add('*')

            # Git commit.
            time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            new_commit = repo.index.commit(message=time + " " + action)
            config.console.info(f"Commit added: [{new_commit.hexsha[0:7]}] {new_commit.message} ")
            config.console.info(f"{new_commit.stats.total['files']} file(s) changed, "
                                f"{new_commit.stats.total['insertions']} insertions(+) "
                                f"{new_commit.stats.total['deletions']} deletions(-).")

            if config.git_mode == "Full":
                # Git push.
                result = repo.remote().push()[0].summary
                config.console.info(f"Push successful: {result}")
        else:
            # git repo is not dirty, we do not have to commit anything.
            config.console.info("Nothing changed, nothing to commit or push.")

    @staticmethod
    def git_pull(config: Config):
        """
        Runs `git pull` on the todotree folder.

        - Does not pull if the previous pull time was less than five minutes ago.
        - Only pulls if git_mode = Full
        """
        if config.git_mode != "Full":
            return
        # Check last pull time.
        # Repo does not have a function to access FETCH_HEAD,
        # So this is done manually.
        fetch_head = config.todo_folder / ".git" / "FETCH_HEAD"

        if fetch_head.exists():
            # Then the repo has been pulled once in its lifetime.
            if datetime.fromtimestamp(fetch_head.stat().st_mtime) < datetime.now() + timedelta(minutes=5):
                # Then the repo is pulled fairly recently. Do not do anything.
                if config.verbose:
                    config.console.info("Repo was pulled recently. Not pulling.")
                return

        # Pull the repo
        config.console.info("Pulling latest changes.")
        config.console.info(Repo(config.todo_folder).remote().pull())



import os
import subprocess
from pathlib import Path

import click

from todotree.Config import Config


# Addons https://click.palletsprojects.com/en/8.1.x/commands/#custom-multi-commands

class Addons:

    def __init__(self, config: Config = Config()):
        self.rv = []
        """List of programs."""

        self.default_config = config
        """Configuration"""

    def list(self):
        """List the addons detected."""
        rv = []
        if Path(self.default_config.addons_folder).exists():
            for filename in Path(self.default_config.addons_folder).iterdir():
                if os.access(filename, os.X_OK):
                    # Then the file is executable, add to the list.
                    rv.append(str(filename.parts[-1]))
            rv.sort()
        self.rv = rv
        return rv

    def run(self, name):
        """Run the addon with the name `name`."""
        file = Path(self.default_config.addons_folder / name)
        if not file.exists():
            raise FileNotFoundError
        result = subprocess.run([file, str(self.default_config.todo_file)], capture_output=True, text=True)
        return result



from typing import Optional, Callable

from pathlib import Path

import xdg_base_dirs as xdg
import yaml

from todotree.ConsolePrefixes import ConsolePrefixes
from todotree.Errors import ConfigFileNotFound


class Config:
    """
    The configuration of todotree.

    For a user guide, please refer to the content in `examples/config.yaml` instead.
    Not all values are mapped 1-to-1.
    """

    def __init__(self):
        #  Main variables.
        self.todo_folder: Path = xdg.xdg_data_home() / "todotree"
        """
        Path to the folder containing the data files.
        
        Relative paths are calculated from the HOME folder. 
        """

        self.project_tree_folder: Path = xdg.xdg_data_home() / "todotree" / "projects"
        """
        Path to the folder containing the projects.
        Defaults to the XDG_DATA_DIR/todotree/projects if not set.
        """

        self.todo_file: Path = Path(self.todo_folder) / "todo.txt"
        """
        Path to the todo.txt file.
        """

        self.done_file: Path = Path(self.todo_folder) / "done.txt"
        """
        Path to the done.txt file.
        """

        self.config_file: Optional[Path] = None
        """
        Path to the config file.
        
        Defaults to None if not set.
        """

        self.addons_folder: Path = Path(self.todo_folder) / "addons"
        """
        Path to the addons folder.
        """

        self.git_mode: str = "None"
        """The mode that git runs in. 
        - None: disables it,
        - Local: add and commits automatically,
        - Full: also pulls and pushes to a remote repo.
        """

        # Features - Enables or disables certain features.
        self.quiet: bool = False
        """A value indicating whether to print anything except the output. Useful in scripts."""

        self.verbose: bool = False
        """A value indicating whether to print more detailed messages."""

        self.enable_wishlist_folder: bool = True
        """A value indicating whether to enable the wishlist folder functionality."""

        self.enable_project_folder: bool = True
        """A value indicating whether to enable the project folder functionality."""

        #  Localization. #
        self.wishlistName: str = "wishlist"
        self.noProjectString: str = "No Project"
        self.noContextString: str = "No Context"
        self.emptyProjectString: str = "> (A) Todo: add next todo for this."

        #  Status Decorators.
        self.console: ConsolePrefixes = ConsolePrefixes(True, ' * ', ' ! ', '!!!')

        # Tree prints.
        self.t_print: str = " ├──"
        self.l_print: str = " │  "
        self.s_print: str = " └──"
        self.e_print: str = "    "

        # Debug
        self.__yaml_object = None
        """Raw config object, for debugging."""

    def __str__(self):
        # Print each attribute in the list.
        s = "Configuration:\n"
        for item in self.__dict__:
            s += f"\t{item} = {self.__dict__[item]} \n"
        return s

    def read(self, path: Optional[Path] = None):
        """
        Loads the configuration from the given file at `path`.

        If empty, reads from default locations.
        """
        if path is not None:
            # print(f"Path is not None: {path}") # FUTURE: Verbose logging
            self.read_from_file(path)
            return
        # xdg compliant directory.
        if Path(xdg.xdg_config_home() / "todotree" / "config.yaml").exists():
            # print(f"Path is XDG Config") # FUTURE: Verbose logging
            self.read_from_file(Path(xdg.xdg_config_home() / "todotree" / "config.yaml"))
            return
        # xdg compliant directory if config file is considered "data".
        if Path(xdg.xdg_data_home() / "todotree" / "config.yaml").exists():
            # print(f"Path is XDG DATA") # FUTURE: Verbose logging
            self.read_from_file(Path(xdg.xdg_data_home() / "todotree" / "config.yaml"))
            return
        # No paths: use the defaults.
        return

    def read_from_file(self, file: Path):
        """Reads and parses yaml content from `file`."""
        self.config_file = file
        try:
            with open(file, 'r') as f:
                self._read_from_yaml(f.read())
        except FileNotFoundError as e:
            raise ConfigFileNotFound from e

    def _read_from_yaml(self, yaml_content: str):
        """Reads and overrides config settings defined in `yaml_content`."""
        # Convert yaml to python object.
        self.__yaml_object = yaml.safe_load(yaml_content)
        if self.__yaml_object is None:
            return
        # Map each item to the self config.
        self.__section('main', self.__section_main)
        self.__section('paths', self.__section_paths)
        self.__section('localization', self.__section_localization)
        self.__section('decorators', self.__section_decorators)
        self.__section('tree', self.__section_tree)

    def __section(self, section_name: str, section_parser: Callable):
        """
        Checks if the section exists and passes it to the sub parser.
        """
        if self.__yaml_object.get(section_name) is not None:
            if self.verbose:
                self.console.info(f"[{section_name}] section found, reading it.")
            # Parse the actual values.
            section_parser()
        elif self.verbose:
            self.console.warning(f"[{section_name}] section not found, skipping it.")

    def __section_tree(self):
        self.t_print = self.__yaml_object['tree'].get('t', self.t_print)
        self.l_print = self.__yaml_object['tree'].get('l', self.l_print)
        self.s_print = self.__yaml_object['tree'].get('s', self.s_print)
        self.e_print = self.__yaml_object['tree'].get('e', self.e_print)

    def __section_decorators(self):
        self.console = ConsolePrefixes(
            self.__yaml_object['decorators'].get('enable_colors', self.console.enable_colors),
            self.__yaml_object['decorators'].get('info', self.console.console_good),
            self.__yaml_object['decorators'].get('warning', self.console.console_warn),
            self.__yaml_object['decorators'].get('error', self.console.console_error))

    def __section_localization(self):
        self.emptyProjectString = self.__yaml_object['localization'].get('empty_project', self.emptyProjectString)
        self.wishlistName = self.__yaml_object['localization'].get('wishlist_name', self.wishlistName)
        self.noProjectString = self.__yaml_object['localization'].get('no_project', self.noProjectString)
        self.noContextString = self.__yaml_object['localization'].get('no_context', self.noContextString)

    def __section_paths(self):
        self.todo_folder = Path(self.__yaml_object['paths'].get('folder', self.todo_folder)).expanduser()
        self.todo_file = Path(
            self.__yaml_object['paths'].get('todo_file', Path(self.todo_folder) / "todo.txt")).expanduser()
        self.done_file = Path(
            self.__yaml_object['paths'].get('done_file', Path(self.todo_folder) / "done.txt")).expanduser()
        self.project_tree_folder = Path(
            self.__yaml_object['paths'].get('project_folder', self.project_tree_folder)).expanduser()
        self.addons_folder = Path(
            self.__yaml_object['paths'].get('addons_folder', self.addons_folder)).expanduser()

    def __section_main(self):
        self.enable_wishlist_folder = self.__yaml_object['main'].get('enable_wishlist_folder',
                                                                     self.enable_wishlist_folder)
        self.enable_project_folder = self.__yaml_object['main'].get('enable_project_folder',
                                                                    self.enable_project_folder)
        self.git_mode = self.__yaml_object['main'].get("git_mode", self.git_mode)
        self.quiet = self.__yaml_object['main'].get("quiet", self.quiet)
        self.verbose = self.__yaml_object['main'].get("verbose", self.verbose)
